import operator
import os
import time
from pathlib import Path

from django.conf import settings
from django.core.files.storage import FileSystemStorage


class FSDataStorage(FileSystemStorage):
    def get_reference_version(self, instance):
        return self._extract_version_ref(instance.geojson.name)

    def make_filename(self, instance):
        root = self._base_path(instance)
        name = "%s_%s.geojson" % (instance.pk, int(time.time() * 1000))
        return root / name

    def _get_names(self, instance):
        root = self._base_path(instance)
        try:
            return self.listdir(root)[1]
        except FileNotFoundError:
            return []

    def list_versions(self, instance):
        names = self._get_names(instance)
        names = [name for name in names if self._is_valid_version(name, instance)]
        versions = [self._version_metadata(name, instance) for name in names]
        versions.sort(reverse=True, key=operator.itemgetter("at"))
        return versions

    def get_version(self, ref, instance):
        with self.open(self.get_version_path(ref, instance), "r") as f:
            return f.read()

    def get_version_path(self, ref, instance):
        base_path = Path(settings.MEDIA_ROOT) / self._base_path(instance)
        fullpath = base_path / f"{instance.pk}_{ref}.geojson"
        if instance.old_id and not fullpath.exists():
            fullpath = base_path / f"{instance.old_id}_{ref}.geojson"
        if not fullpath.exists():
            raise ValueError(f"Invalid version reference: {ref}")
        return fullpath

    def onDatalayerSave(self, instance):
        self.purge_gzip(instance)
        self.purge_old_versions(instance, keep=settings.UMAP_KEEP_VERSIONS)

    def onDatalayerDelete(self, instance):
        self.purge_gzip(instance)
        self.purge_old_versions(instance, keep=None)

    def _extract_version_ref(self, path):
        version = path.split(".")[0]
        if "_" in version:
            return version.split("_")[-1]
        return version

    def _base_path(self, instance):
        path = ["datalayer", str(instance.map.pk)[-1]]
        if len(str(instance.map.pk)) > 1:
            path.append(str(instance.map.pk)[-2])
        path.append(str(instance.map.pk))
        return Path(os.path.join(*path))

    def _is_valid_version(self, name, instance):
        valid_prefixes = [name.startswith("%s_" % instance.pk)]
        if instance.old_id:
            valid_prefixes.append(name.startswith("%s_" % instance.old_id))
        return any(valid_prefixes) and name.endswith(".geojson")

    def _version_metadata(self, name, instance):
        ref = self._extract_version_ref(name)
        return {
            "name": name,
            "ref": ref,
            "at": ref,
            "size": self.size(self._base_path(instance) / name),
        }

    def purge_old_versions(self, instance, keep=None):
        root = self._base_path(instance)
        versions = self.list_versions(instance)
        if keep is not None:
            versions = versions[keep:]
        deleted = 0
        for version in versions:
            name = version["name"]
            # Should not be in the list, but ensure to not delete the file
            # currently used in database
            if keep is not None and instance.geojson.name.endswith(name):
                continue
            try:
                self.delete(root / name)
            except FileNotFoundError:
                pass
            else:
                deleted += 1
        return deleted

    def purge_gzip(self, instance):
        root = self._base_path(instance)
        names = self._get_names(instance)
        prefixes = [f"{instance.pk}_"]
        if instance.old_id:
            prefixes.append(f"{instance.old_id}_")
        prefixes = tuple(prefixes)
        for name in names:
            if name.startswith(prefixes) and name.endswith(".gz"):
                self.delete(root / name)



import pytest
from playwright.sync_api import expect

from ..base import DataLayerFactory

pytestmark = pytest.mark.django_db

OSM_DATA = {
    "type": "FeatureCollection",
    "features": [
        {
            "type": "Feature",
            "geometry": {"type": "Point", "coordinates": [2.49, 48.79]},
            "properties": {
                "amenity": "restaurant",
                "cuisine": "italian",
                "name": "A Casa di Nonna",
                "panoramax": "d811b398-d930-4cf8-95a2-0c29c34d9fca",
                "phone": "+33 1 48 89 54 12",
                "takeaway:covid19": "yes",
                "wheelchair": "no",
                "id": "node/1130849864",
            },
            "id": "AzMjk",
        },
    ],
    "_umap_options": {
        "popupTemplate": "OSM",
    },
}


def test_openstreetmap_popup(live_server, map, page):
    DataLayerFactory(map=map, data=OSM_DATA)
    page.goto(f"{live_server.url}{map.get_absolute_url()}#18/48.79/2.49")
    expect(page.locator(".umap-icon-active")).to_be_hidden()
    page.locator(".leaflet-marker-icon").click()
    expect(page.get_by_role("heading", name="A Casa di Nonna")).to_be_visible()
    expect(page.get_by_text("+33 1 48 89 54 12")).to_be_visible()
    img = page.locator(".umap-popup-content img")
    expect(img).to_have_attribute(
        "src",
        "https://api.panoramax.xyz/api/pictures/d811b398-d930-4cf8-95a2-0c29c34d9fca/sd.jpg",
    )



import asyncio
import logging

import redis.asyncio as redis
from django.conf import settings
from django.core.signing import TimestampSigner
from django.urls import path
from pydantic import ValidationError

from .payloads import (
    JoinRequest,
    JoinResponse,
    ListPeersResponse,
    OperationMessage,
    PeerMessage,
    Request,
    SavedMessage,
)


async def application(scope, receive, send):
    path = scope["path"].lstrip("/")
    for pattern in urlpatterns:
        if matched := pattern.resolve(path):
            await matched.func(scope, receive, send, **matched.kwargs)
            break
    else:
        await send({"type": "websocket.close"})


async def sync(scope, receive, send, **kwargs):
    peer = Peer(kwargs["map_id"])
    peer._send = send
    while True:
        event = await receive()

        if event["type"] == "websocket.connect":
            try:
                await peer.connect()
                await send({"type": "websocket.accept"})
            except ValueError:
                await send({"type": "websocket.close"})

        if event["type"] == "websocket.disconnect":
            await peer.disconnect()
            break

        if event["type"] == "websocket.receive":
            if event["text"] == "ping":
                await send({"type": "websocket.send", "text": "pong"})
            else:
                await peer.receive(event["text"])


class Peer:
    def __init__(self, map_id, username=None):
        self.username = username or ""
        self.map_id = map_id
        self.is_authenticated = False
        self._subscriptions = []

    @property
    def room_key(self):
        return f"umap:{self.map_id}"

    @property
    def peer_key(self):
        return f"user:{self.map_id}:{self.peer_id}"

    async def get_peers(self):
        known = await self.client.hgetall(self.room_key)
        active = await self.client.pubsub_channels(f"user:{self.map_id}:*")
        if not active:
            # Poor man way of deleting stale usernames from the store
            # HEXPIRE command is not in the open source Redis version
            await self.client.delete(self.room_key)
            await self.store_username()
        active = [name.split(b":")[-1] for name in active]
        if self.peer_id.encode() not in active:
            # Our connection may not yet be active
            active.append(self.peer_id.encode())
        return {k: v for k, v in known.items() if k in active}

    async def store_username(self):
        await self.client.hset(self.room_key, self.peer_id, self.username)

    async def listen_to_channel(self, channel_name):
        async def reader(pubsub):
            await pubsub.subscribe(channel_name)
            while True:
                if pubsub.connection is None:
                    # It has been unsubscribed/closed.
                    break
                try:
                    message = await pubsub.get_message(ignore_subscribe_messages=True)
                except Exception as err:
                    logging.debug(err)
                    break
                if message is not None:
                    await self.send(message["data"].decode())
                await asyncio.sleep(0.001)  # Be nice with the server

        async with self.client.pubsub() as pubsub:
            self._subscriptions.append(pubsub)
            asyncio.create_task(reader(pubsub))

    async def listen(self):
        await self.listen_to_channel(self.room_key)
        await self.listen_to_channel(self.peer_key)

    async def connect(self):
        self.client = redis.from_url(settings.REDIS_URL)

    async def disconnect(self):
        if self.is_authenticated:
            await self.client.hdel(self.room_key, self.peer_id)
            for pubsub in self._subscriptions:
                await pubsub.unsubscribe()
                await pubsub.close()
            await self.send_peers_list()
        await self.client.aclose()

    async def send_peers_list(self):
        message = ListPeersResponse(peers=await self.get_peers())
        await self.broadcast(message.model_dump_json())

    async def broadcast(self, message):
        logging.debug("BROADCASTING", message)
        # Send to all channels (including sender!)
        await self.client.publish(self.room_key, message)

    async def send_to(self, peer_id, message):
        logging.debug("SEND TO", peer_id, message)
        # Send to one given channel
        await self.client.publish(f"user:{self.map_id}:{peer_id}", message)

    async def receive(self, text_data):
        if not self.is_authenticated:
            logging.debug("AUTHENTICATING", text_data)
            message = JoinRequest.model_validate_json(text_data)
            signed = TimestampSigner().unsign_object(message.token, max_age=30)
            user, map_id, permissions = signed.values()
            assert str(map_id) == self.map_id
            if "edit" not in permissions:
                return await self.disconnect()
            self.peer_id = message.peer
            self.username = message.username
            logging.debug("AUTHENTICATED", self.peer_id)
            await self.store_username()
            await self.listen()
            response = JoinResponse(peer=self.peer_id, peers=await self.get_peers())
            await self.send(response.model_dump_json())
            await self.send_peers_list()
            self.is_authenticated = True
            return

        try:
            incoming = Request.model_validate_json(text_data)
        except ValidationError as error:
            message = (
                f"An error occurred when receiving the following message: {text_data!r}"
            )
            logging.error(message, error)
        else:
            match incoming.root:
                # Broadcast all operation messages to connected peers
                case OperationMessage():
                    await self.broadcast(text_data)

                # Broadcast the new map state to connected peers
                case SavedMessage():
                    await self.broadcast(text_data)

                # Send peer messages to the proper peer
                case PeerMessage():
                    await self.send_to(incoming.root.recipient, text_data)

    async def send(self, text):
        logging.debug("  FORWARDING TO", self.peer_id, text)
        try:
            await self._send({"type": "websocket.send", "text": text})
        except Exception as err:
            logging.debug("Error sending message:", text)
            logging.debug(err)


urlpatterns = [path("ws/sync/<str:map_id>", name="ws_sync", view=sync)]



import json
import sys

from django.core.management.base import BaseCommand
from django.db import connection
from psycopg.types.json import Jsonb

from umap.models import Map, TileLayer


class Command(BaseCommand):
    help = """Clean tilelayer in database

    This will simply replace the URL in maps settings:
    umap clean_tilelayer http://my.old/url/template http://my.new/url/template

    This will replace the whole tilelayer in maps settings by the one with this name:
    umap clean_tilelayer http://my.old/url/template "some string"

    This will replace the whole tilelayer in maps settings by the one with this id:
    umap clean_tilelayer http://my.old/url/template an_id

    This will delete the whole tilelayer from maps settings:
    umap clean_tilelayer http://my.old/url/template

    To get the available tilelayers in db (available for users):
    umap clean_tilelayer --available

    To get statistics of tilelayers usage in db (including custom ones):
    umap clean_tilelayer --available
    """

    def add_arguments(self, parser):
        parser.add_argument("old", nargs="?", help="url template we want to clean")
        parser.add_argument(
            "new", help="what to replace this tilelayer with", nargs="?"
        )
        parser.add_argument(
            "--no-input", action="store_true", help="Do not ask for confirm."
        )
        parser.add_argument(
            "--available", action="store_true", help="List known tilelayers."
        )
        parser.add_argument(
            "--stats", action="store_true", help="Display stats on tilelayer usage."
        )

    def handle(self, *args, **options):
        self.no_input = options["no_input"]
        if options["available"]:
            self.list_available()
            sys.exit()
        if options["stats"]:
            self.stats()
            sys.exit()
        old = options["old"]
        new = options["new"]
        if not old:
            sys.exit("⚠ You must define an url_template")

        count = Map.objects.filter(
            settings__properties__tilelayer__url_template=old
        ).count()
        if not count:
            self.stdout.write("⚠ No map found. Exiting.")
            sys.exit()
        self.stdout.write(f"{count} maps found.")
        if not new:
            self.delete(old)
        elif new.startswith("http"):
            self.replace_url(old, new)
        else:
            # Let's consider it's a name or an id
            self.replace_tilelayer(old, new)

    def confirm(self, message):
        if self.no_input:
            return True
        result = input("%s (y/N) " % message) or "n"
        if not result[0].lower() == "y":
            self.stdout.write("⚠ Action cancelled.")
            sys.exit()
        return True

    def delete(self, old):
        if self.confirm(
            "Are you sure you want to delete the tilelayer key from all those "
            "maps settings ?"
        ):
            with connection.cursor() as cursor:
                ret = cursor.execute(
                    "UPDATE umap_map "
                    "SET settings['properties'] = (settings->'properties') - 'tilelayer'"
                    "WHERE settings->'properties'->'tilelayer'->'url_template' = %s",
                    [Jsonb(old)],
                )
            self.stdout.write(f"✔ Deleted {old} from {ret.rowcount} maps.")

    def replace_url(self, old, new):
        if self.confirm(
            f"Are you sure you want to replace '{old}'' by '{new}'' from all those "
            "map settings ?"
        ):
            with connection.cursor() as cursor:
                ret = cursor.execute(
                    "UPDATE umap_map "
                    "SET settings['properties']['tilelayer']['url_template'] = %s "
                    "WHERE settings->'properties'->'tilelayer'->'url_template' = %s",
                    [Jsonb(new), Jsonb(old)],
                )
            self.stdout.write(f"✔ Replaced {old} by {new} in {ret.rowcount} maps.")

    def replace_tilelayer(self, old, new):
        try:
            tilelayer = TileLayer.objects.get(name=new)
        except TileLayer.DoesNotExist:
            try:
                tilelayer = TileLayer.objects.get(id=new)
            except (TileLayer.DoesNotExist, ValueError):
                sys.exit(f"⚠ Cannot find a TileLayer with name or id = '{new}'.")
        if self.confirm(
            f"Are you sure you want to replace {old} by '{tilelayer.name}' "
            "from all those map settings ?"
        ):
            with connection.cursor() as cursor:
                ret = cursor.execute(
                    "UPDATE umap_map "
                    "SET settings['properties']['tilelayer'] = %s "
                    "WHERE settings->'properties'->'tilelayer'->'url_template' = %s",
                    [Jsonb(tilelayer.json), Jsonb(old)],
                )
            self.stdout.write(
                f"✔ Replaced {old} by {tilelayer.name} in {ret.rowcount} maps."
            )

    def list_available(self):
        tilelayers = TileLayer.objects.all()
        for tilelayer in tilelayers:
            print(f"{tilelayer.pk} '{tilelayer.name}' {tilelayer.url_template}")

    def stats(self):
        with connection.cursor() as cursor:
            cursor.execute(
                "SELECT COUNT(*) as count, "
                "settings->'properties'->'tilelayer'->'url_template' as url "
                "FROM umap_map "
                "GROUP BY settings->'properties'->'tilelayer'->'url_template' "
                "ORDER BY count DESC"
            )
            res = cursor.fetchall()
        for count, url in res:
            print(f"{count}\t{url}")



from datetime import datetime, timedelta

from django.core.management.base import BaseCommand

from umap.models import DataLayer, Map


class Command(BaseCommand):
    help = "Remove maps in trash. Eg.: umap empty_trash --days 7"

    def add_arguments(self, parser):
        parser.add_argument(
            "--days",
            help="Number of days to consider maps for removal",
            default=30,
            type=int,
        )
        parser.add_argument(
            "--dry-run",
            help="Pretend to delete but just report",
            action="store_true",
        )

    def handle(self, *args, **options):
        days = options["days"]
        since = (datetime.utcnow() - timedelta(days=days)).date()
        print(f"Deleting map in trash since {since}")
        maps = Map.objects.filter(share_status=Map.DELETED, modified_at__lt=since)
        for map in maps:
            map_id = map.id
            map_name = map.name
            trashed_at = map.modified_at.date()
            if not options["dry_run"]:
                map.delete()
            print(f"Deleted map {map_name} ({map_id}), trashed at {trashed_at}")
        print(f"Deleting layers in trash since {since}")
        layers = DataLayer.objects.filter(
            share_status=DataLayer.DELETED, modified_at__lt=since
        )
        for layer in layers:
            layer_id = layer.uuid
            layer_name = layer.name
            trashed_at = layer.modified_at.date()
            if not options["dry_run"]:
                layer.delete()
            print(f"Deleted layer {layer_name} ({layer_id}), trashed at {trashed_at}")



import pytest

from .base import TileLayerFactory

pytestmark = pytest.mark.django_db


def test_tilelayer_json():
    tilelayer = TileLayerFactory(
        attribution="Attribution",
        maxZoom=19,
        minZoom=0,
        name="Name",
        rank=1,
        tms=True,
        url_template="http://{s}.x.fr/{z}/{x}/{y}",
    )
    assert tilelayer.json == {
        "attribution": "Attribution",
        "id": tilelayer.id,
        "maxZoom": 19,
        "minZoom": 0,
        "name": "Name",
        "rank": 1,
        "tms": True,
        "url_template": "http://{s}.x.fr/{z}/{x}/{y}",
    }



from agnocomplete.core import AgnocompleteModel
from agnocomplete.register import register
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models.functions import Length


@register
class AutocompleteUser(AgnocompleteModel):
    model = get_user_model()
    fields = settings.USER_AUTOCOMPLETE_FIELDS

    def item(self, current_item):
        data = super().item(current_item)
        data["url"] = current_item.get_url()
        return data

    def build_extra_filtered_queryset(self, queryset, **kwargs):
        order_by = []
        for field_name in self.fields:
            if not field_name[0].isalnum():
                field_name = field_name[1:]
            order_by.append(Length(field_name).asc())
        return queryset.order_by(*order_by)



import pytest
from playwright.sync_api import expect

pytestmark = pytest.mark.django_db


def test_reuse_template_button(map, datalayer, page, live_server, context):
    map.is_template = True
    map.save()
    page.goto(f"{live_server.url}{map.get_absolute_url()}")
    expect(page.get_by_text("Reuse this template")).to_be_visible()
    with context.expect_page() as new_page_info:
        page.get_by_text("Reuse this template").click()
    new_page = new_page_info.value
    expect(new_page.get_by_text("Reuse this template")).to_be_hidden()
    expect(page.locator(".leaflet-marker-icon")).to_have_count(1)


def test_load_template_from_panel(map, datalayer, page, live_server):
    map.name = "My Great Template"
    map.is_template = True
    map.save()
    page.goto(f"{live_server.url}/en/map/new")
    page.get_by_title("Load template").click()
    page.get_by_role("button", name="From community").click()
    page.get_by_label("My Great Template").check()
    page.get_by_role("button", name="Load template with data").click()
    expect(page.get_by_text("My Great Template")).to_be_visible()
    expect(page.locator(".leaflet-marker-icon")).to_have_count(1)


def test_load_template_without_data(map, datalayer, page, live_server):
    map.name = "My Great Template"
    map.is_template = True
    map.save()
    page.goto(f"{live_server.url}/en/map/new")
    page.get_by_title("Load template").click()
    page.get_by_role("button", name="From community").click()
    page.get_by_label("My Great Template").check()
    page.locator(".panel").get_by_role(
        "button", name="Load template", exact=True
    ).click()
    expect(page.get_by_text("My Great Template")).to_be_visible()
    expect(page.locator(".leaflet-marker-icon")).to_have_count(0)



import json
import re
from pathlib import Path

from umap.models import DataLayer


def save_and_get_json(page):
    with page.expect_response(re.compile(r".*/datalayer/create/.*")):
        page.get_by_role("button", name="Save").click()
    datalayer = DataLayer.objects.last()
    return json.loads(Path(datalayer.geojson.path).read_text())



# Generated by Django 5.0.2 on 2024-04-17 09:18

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0019_migrate_internal_remote_datalayers"),
    ]

    operations = [
        migrations.AlterField(
            model_name="tilelayer",
            name="url_template",
            field=models.CharField(
                help_text="URL template using OSM tile format", max_length=400
            ),
        ),
    ]



import json
from pathlib import Path


def test_ids_generation(page, live_server, tilelayer):
    page.goto(f"{live_server.url}/en/map/new/")

    # Click on the Draw a line button on a new map.
    create_polyline = page.locator(".umap-edit-bar ").get_by_title("Draw a polyline")
    create_polyline.click()

    map = page.locator("#map")
    map.click(position={"x": 200, "y": 200})
    map.click(position={"x": 100, "y": 100})
    # Click again to finish
    map.click(position={"x": 100, "y": 100})

    # Click on the Draw a polygon button on a new map.
    create_polygon = page.locator(".umap-edit-bar ").get_by_title("Draw a polygon")
    create_polygon.click()

    map = page.locator("#map")
    map.click(position={"x": 300, "y": 300})
    map.click(position={"x": 300, "y": 400})
    map.click(position={"x": 350, "y": 450})
    # Click again to finish
    map.click(position={"x": 350, "y": 450})

    download_panel = page.get_by_title("Share and download")
    download_panel.click()

    button = page.get_by_role("button", name="geojson")

    with page.expect_download() as download_info:
        button.click()

    download = download_info.value

    path = Path("/tmp/") / download.suggested_filename
    download.save_as(path)
    downloaded = json.loads(path.read_text())

    assert "features" in downloaded
    features = downloaded["features"]
    assert len(features) == 2
    assert "id" in features[0]
    assert "id" in features[1]



import re

import pytest
from playwright.sync_api import expect

from umap.models import Star

pytestmark = pytest.mark.django_db


def test_star_button_is_active_if_logged_in(map, live_server, page, login, user):
    login(user)
    assert not Star.objects.count()
    page.goto(f"{live_server.url}{map.get_absolute_url()}")
    page.get_by_title("About").click()
    button = page.locator(".icon-star")
    expect(button).to_be_visible()
    with page.expect_response(re.compile(".*/star/")):
        button.click()
    expect(button).to_be_hidden()
    # Button has changed
    expect(page.locator(".icon-starred")).to_be_visible()
    assert Star.objects.count() == 1


def test_star_button_inctive_if_not_logged_in(map, live_server, page):
    page.goto(f"{live_server.url}{map.get_absolute_url()}")
    page.get_by_title("About").click()
    button = page.locator(".icon-star")
    button.click()
    expect(page.get_by_text("You must be logged in")).to_be_visible()



from django.core.management.base import BaseCommand

from umap.models import Map, User


class Command(BaseCommand):
    help = "Replace a user by another as owner or editor or team member. Eg. umap switch_user oldUserName newUserName"

    def add_arguments(self, parser):
        parser.add_argument("old", help="Old username.")
        parser.add_argument("new", help="New username.")
        parser.add_argument(
            "--dry-run",
            help="Do not replace for real, just display actions",
            action="store_true",
        )
        parser.add_argument(
            "--delete-user",
            help="Also delete old user",
            action="store_true",
        )

    def handle(self, *args, **options):
        old = User.objects.get(username=options["old"])
        new = User.objects.get(username=options["new"])
        print(f"Replacing usre {old} by user {new}")
        owned_maps = old.owned_maps.all()
        print(f"Replacing owner of {len(owned_maps)} maps")
        if not options["dry_run"]:
            for mm in owned_maps:
                mm.owner = new
                mm.save()

        as_editor = Map.objects.filter(editors=old)
        print(f"Replacing editor of {len(as_editor)} maps")
        if not options["dry_run"]:
            for mm in as_editor:
                mm.editors.remove(old)
                mm.editors.add(new)
                mm.save()

        teams = old.teams.all()
        print(f"Replacing editor of {len(teams)} maps")
        if not options["dry_run"]:
            for team in teams:
                team.users.remove(old)
                team.users.add(new)
                team.save()

        if not options["dry_run"] and options["delete_user"]:
            old.delete()
            print("Deleted old user.")



import re

import pytest
import redis
from django.conf import settings
from playwright.sync_api import expect

from umap.models import DataLayer, Map

from ..base import DataLayerFactory, MapFactory

DATALAYER_UPDATE = re.compile(r".*/datalayer/update/.*")

pytestmark = pytest.mark.django_db


def setup_function():
    # Sync client to prevent headache with pytest / pytest-asyncio and async
    client = redis.from_url(settings.REDIS_URL)
    # Make sure there are no dead peers in the Redis hash, otherwise asking for
    # operations from another peer may never be answered
    # FIXME this should not happen in an ideal world
    assert client.connection_pool.connection_kwargs["db"] == 15
    client.flushdb()


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_markers(new_page, asgi_live_server, tilelayer):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page("Page B")
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    a_marker_pane = peerA.locator(".leaflet-marker-pane > div")
    b_marker_pane = peerB.locator(".leaflet-marker-pane > div")
    expect(a_marker_pane).to_have_count(0)
    expect(b_marker_pane).to_have_count(0)

    # Add a marker from peer A
    a_create_marker = peerA.get_by_title("Draw a marker")
    expect(a_create_marker).to_be_visible()
    a_create_marker.click()

    a_map_el = peerA.locator("#map")
    a_map_el.click(position={"x": 220, "y": 220})
    expect(a_marker_pane).to_have_count(1)
    expect(b_marker_pane).to_have_count(1)
    # Peer B should not be in state dirty
    expect(peerB.get_by_role("button", name="View")).to_be_visible()
    expect(peerB.get_by_role("button", name="Cancel edits")).to_be_hidden()
    peerA.locator("body").type("Synced name")
    peerA.locator("body").press("Escape")
    peerA.wait_for_timeout(300)

    peerB.locator(".leaflet-marker-icon").first.click()
    peerB.get_by_role("button", name="Toggle edit mode (⇧+Click)").click()
    expect(peerB.locator('input[name="name"]')).to_have_value("Synced name")

    a_first_marker = peerA.locator("div:nth-child(4) > div:nth-child(2)").first
    b_first_marker = peerB.locator("div:nth-child(4) > div:nth-child(2)").first

    # Add a second marker from peer B
    b_create_marker = peerB.get_by_title("Draw a marker")
    expect(b_create_marker).to_be_visible()
    b_create_marker.click()

    b_map_el = peerB.locator("#map")
    b_map_el.click(position={"x": 225, "y": 225})
    expect(a_marker_pane).to_have_count(2)
    expect(b_marker_pane).to_have_count(2)

    # Drag a marker on peer B and check that it moved on peer A
    assert a_first_marker.bounding_box() == b_first_marker.bounding_box()
    b_old_bbox = b_first_marker.bounding_box()
    b_first_marker.drag_to(b_map_el, target_position={"x": 250, "y": 250})

    assert b_old_bbox is not b_first_marker.bounding_box()
    assert a_first_marker.bounding_box() == b_first_marker.bounding_box()

    # Delete a marker from peer A and check it's been deleted on peer B
    a_first_marker.click(button="right")
    peerA.get_by_role("button", name="Delete this feature").click()
    expect(a_marker_pane).to_have_count(1)
    expect(b_marker_pane).to_have_count(1)


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_polygons(context, asgi_live_server, tilelayer):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = context.new_page()
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = context.new_page()
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    b_map_el = peerB.locator("#map")

    # Click on the Draw a polygon button on a new map.
    create_line = peerA.locator(".umap-edit-bar ").get_by_title("Draw a polygon")
    create_line.click()

    a_polygons = peerA.locator(".leaflet-overlay-pane path[fill='DarkBlue']")
    b_polygons = peerB.locator(".leaflet-overlay-pane path[fill='DarkBlue']")
    expect(a_polygons).to_have_count(0)
    expect(b_polygons).to_have_count(0)

    # Click on the map, it will create a polygon.
    map = peerA.locator("#map")
    map.click(position={"x": 200, "y": 200})
    map.click(position={"x": 100, "y": 200})
    map.click(position={"x": 100, "y": 100})
    map.click(position={"x": 100, "y": 100})

    # It is created on peerA, and should be on peerB
    expect(a_polygons).to_have_count(1)
    expect(b_polygons).to_have_count(1)

    # Escaping the edition should not duplicate
    peerA.keyboard.press("Escape")
    expect(a_polygons).to_have_count(1)
    expect(b_polygons).to_have_count(1)

    # change the geometry by moving a point on peer B
    a_polygon = peerA.locator("path")
    b_polygon = peerB.locator("path")
    b_polygon_bbox_t1 = b_polygon.bounding_box()
    a_polygon_bbox_t1 = a_polygon.bounding_box()
    assert b_polygon_bbox_t1 == a_polygon_bbox_t1

    b_polygon.click()
    peerB.get_by_role("button", name="Toggle edit mode (⇧+Click)").click()

    edited_vertex = peerB.locator("div:nth-child(6)").first
    edited_vertex.drag_to(b_map_el, target_position={"x": 233, "y": 126})
    peerB.keyboard.press("Escape")

    b_polygon_bbox_t2 = b_polygon.bounding_box()
    a_polygon_bbox_t2 = a_polygon.bounding_box()

    assert b_polygon_bbox_t2 != b_polygon_bbox_t1
    assert b_polygon_bbox_t2 == a_polygon_bbox_t2

    # Move the polygon on peer B and check it moved also on peer A
    b_polygon.click()
    peerB.get_by_role("button", name="Toggle edit mode (⇧+Click)").click()

    b_polygon.drag_to(b_map_el, target_position={"x": 400, "y": 400})
    peerB.keyboard.press("Escape")
    b_polygon_bbox_t3 = b_polygon.bounding_box()
    a_polygon_bbox_t3 = a_polygon.bounding_box()

    assert b_polygon_bbox_t3 != b_polygon_bbox_t2
    assert b_polygon_bbox_t3 == a_polygon_bbox_t3

    # Delete a polygon from peer A and check it's been deleted on peer B
    a_polygon.click(button="right")
    peerA.get_by_role("button", name="Delete this feature").click()
    expect(a_polygons).to_have_count(0)
    expect(b_polygons).to_have_count(0)


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_map_properties(
    new_page, asgi_live_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = new_page()
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page()
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    # Name change is synced
    peerA.get_by_role("button", name="Edit map name and caption").click()
    peerA.locator('input[name="name"]').click()
    peerA.locator('input[name="name"]').fill("it syncs!")

    expect(peerB.locator(".map-name").last).to_have_text("it syncs!")

    # Zoom control is synced
    peerB.get_by_role("button", name="Map advanced properties").click()
    peerB.locator("summary").filter(has_text="User interface options").click()
    switch = peerB.locator("div.formbox").filter(
        has_text=re.compile("Display the zoom control")
    )
    expect(switch).to_be_visible()
    switch.get_by_text("Never").click()

    expect(peerA.locator(".leaflet-control-zoom")).to_be_hidden()


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_datalayer_properties(
    new_page, asgi_live_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = new_page()
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page()
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    # Layer addition, name and type are synced
    peerA.get_by_role("button", name="Manage layers").click()
    peerA.get_by_role("button", name="Add a layer").click()
    peerA.locator('input[name="name"]').click()
    peerA.locator('input[name="name"]').fill("synced layer!")
    peerA.get_by_role("combobox").select_option("Choropleth")
    peerA.locator("body").press("Escape")

    peerB.get_by_role("button", name="Manage layers").click()
    peerB.locator(".panel.right").get_by_role("button", name="Edit").first.click()
    expect(peerB.locator('input[name="name"]')).to_have_value("synced layer!")
    expect(peerB.get_by_role("combobox")).to_have_value("Choropleth")


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_cloned_polygons(
    context, asgi_live_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create two tabs
    peerA = context.new_page()
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = context.new_page()
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    b_map_el = peerB.locator("#map")

    # Click on the Draw a polygon button on a new map.
    create_line = peerA.locator(".umap-edit-bar ").get_by_title("Draw a polygon")
    create_line.click()

    a_polygons = peerA.locator(".leaflet-overlay-pane path[fill='DarkBlue']")
    b_polygons = peerB.locator(".leaflet-overlay-pane path[fill='DarkBlue']")
    expect(a_polygons).to_have_count(0)
    expect(b_polygons).to_have_count(0)

    # Click on the map, it will create a polygon.
    map = peerA.locator("#map")
    map.click(position={"x": 200, "y": 200})
    map.click(position={"x": 100, "y": 200})
    map.click(position={"x": 100, "y": 100})
    map.click(position={"x": 200, "y": 100})
    map.click(position={"x": 200, "y": 100})

    # Escaping the edition syncs
    peerA.keyboard.press("Escape")
    expect(a_polygons).to_have_count(1)
    expect(b_polygons).to_have_count(1)

    # Save from peer A
    peerA.get_by_role("button", name="Save").click()

    b_polygon = peerB.locator("path")

    # Clone on peer B and save
    b_polygon.click(button="right", delay=200)
    peerB.get_by_role("button", name="Clone this feature").click()

    expect(peerB.locator("path")).to_have_count(2)

    peerB.locator("path").nth(1).drag_to(b_map_el, target_position={"x": 400, "y": 400})
    peerB.locator("path").nth(1).click()
    peerB.locator("summary").filter(has_text="Shape properties").click()
    peerB.locator(".umap-field-color button.define").first.click()
    peerB.get_by_title("Orchid", exact=True).first.click()
    peerB.locator("#map").press("Escape")
    peerB.get_by_role("button", name="Save").click()

    expect(peerB.locator("path")).to_have_count(2)


@pytest.mark.xdist_group(name="websockets")
def test_websocket_connection_can_sync_late_joining_peer(
    new_page, asgi_live_server, tilelayer
):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    DataLayerFactory(map=map, data={})

    # Create first peer (A) and have it join immediately
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    # Add a marker from peer A
    a_create_marker = peerA.get_by_title("Draw a marker")
    expect(a_create_marker).to_be_visible()
    a_create_marker.click()

    a_map_el = peerA.locator("#map")
    a_map_el.click(position={"x": 220, "y": 220})
    peerA.locator("body").type("First marker")
    peerA.locator("body").press("Escape")
    peerA.wait_for_timeout(300)

    # Add a polygon from peer A
    create_polygon = peerA.locator(".umap-edit-bar ").get_by_title("Draw a polygon")
    create_polygon.click()

    a_map_el.click(position={"x": 200, "y": 200})
    a_map_el.click(position={"x": 100, "y": 200})
    a_map_el.click(position={"x": 100, "y": 100})
    a_map_el.click(position={"x": 200, "y": 100})
    a_map_el.click(position={"x": 200, "y": 100})
    peerA.keyboard.press("Escape")

    # Now create peer B and have it join
    peerB = new_page("Page B")
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    # Check if peer B has received all the updates
    b_marker_pane = peerB.locator(".leaflet-marker-pane > div")
    b_polygons = peerB.locator(".leaflet-overlay-pane path[fill='DarkBlue']")

    expect(b_marker_pane).to_have_count(1)
    expect(b_polygons).to_have_count(1)

    # Verify marker properties
    peerB.locator(".leaflet-marker-icon").first.click()
    peerB.get_by_role("button", name="Toggle edit mode (⇧+Click)").click()
    expect(peerB.locator('input[name="name"]')).to_have_value("First marker")

    # Verify polygon exists (we've already checked the count)
    b_polygon = peerB.locator("path")
    expect(b_polygon).to_be_visible()

    # Optional: Verify polygon properties if you have any specific ones set

    # Clean up: close edit mode
    peerB.locator("body").press("Escape")


@pytest.mark.xdist_group(name="websockets")
def test_should_sync_datalayers(new_page, asgi_live_server, tilelayer):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()

    assert not DataLayer.objects.count()

    # Create two tabs
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page("Page B")
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    # Create a new layer from peerA
    peerA.get_by_role("button", name="Manage layers").click()
    peerA.get_by_role("button", name="Add a layer").click()

    # Check layer has been sync to peerB
    peerB.get_by_role("button", name="Open browser").click()
    expect(peerB.get_by_text("Layer 1")).to_be_visible()

    # Draw a marker in layer 1 from peerA
    peerA.get_by_role("button", name="Draw a marker (Ctrl+M)").click()
    peerA.locator("#map").click()

    # Check marker is visible from peerB
    expect(peerB.locator(".leaflet-marker-icon")).to_be_visible()

    # Save layer to the server
    with peerA.expect_response(re.compile(".*/datalayer/create/.*")):
        peerA.get_by_role("button", name="Save").click()

    assert DataLayer.objects.count() == 1

    # Create another layer from peerA and draw a marker on it (without saving to server)
    peerA.get_by_role("button", name="Manage layers").click()
    peerA.get_by_role("button", name="Add a layer").click()
    peerA.get_by_role("button", name="Draw a marker (Ctrl+M)").click()
    peerA.locator("#map").click()

    # Make sure this new marker is in Layer 2 for peerB
    # Show features for this layer in the brower.
    peerB.locator("summary").filter(has_text="Layer 2").click()
    expect(peerB.locator("li").filter(has_text="Layer 2")).to_be_visible()
    peerB.locator(".panel.left").get_by_role("button", name="Show/hide layer").nth(
        1
    ).click()
    expect(peerB.locator(".leaflet-marker-icon")).to_be_visible()

    # Now draw a marker from peerB
    peerB.get_by_role("button", name="Draw a marker (Ctrl+M)").click()
    peerB.locator("#map").click()
    peerB.locator('input[name="name"]').fill("marker from peerB")

    # Save from peer B
    with peerB.expect_response(re.compile(".*/datalayer/create/.*")):
        peerB.get_by_role("button", name="Save").click()

    assert DataLayer.objects.count() == 2

    # Check this new marker is visible from peerA
    peerA.get_by_role("button", name="Open browser").click()
    peerA.locator(".panel.left").get_by_role("button", name="Show/hide layer").nth(
        1
    ).click()

    # Peer A should not be in dirty state
    expect(peerA.locator("body")).not_to_have_class(re.compile(".*umap-is-dirty.*"))

    # Peer A should only have two markers
    expect(peerA.locator(".leaflet-marker-icon")).to_have_count(2)

    assert DataLayer.objects.count() == 2


@pytest.mark.xdist_group(name="websockets")
def test_should_sync_datalayers_delete(new_page, asgi_live_server, tilelayer):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()
    data1 = {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "properties": {
                    "name": "Point 1",
                },
                "geometry": {"type": "Point", "coordinates": [0.065918, 48.385442]},
            },
        ],
        "_umap_options": {
            "name": "datalayer 1",
        },
    }
    data2 = {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "properties": {
                    "name": "Point 2",
                },
                "geometry": {"type": "Point", "coordinates": [3.55957, 49.767074]},
            },
        ],
        "_umap_options": {
            "name": "datalayer 2",
        },
    }
    DataLayerFactory(map=map, data=data1)
    DataLayerFactory(map=map, data=data2)

    # Create two tabs
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page("Page B")
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    peerA.get_by_role("button", name="Open browser").click()
    expect(peerA.locator(".panel").get_by_text("datalayer 1")).to_be_visible()
    expect(peerA.locator(".panel").get_by_text("datalayer 2")).to_be_visible()
    peerB.get_by_role("button", name="Open browser").click()
    expect(peerB.locator(".panel").get_by_text("datalayer 1")).to_be_visible()
    expect(peerB.locator(".panel").get_by_text("datalayer 2")).to_be_visible()

    # Delete "datalayer 2" in peerA
    peerA.locator(".datalayer").get_by_role("button", name="Delete layer").first.click()
    expect(peerA.locator(".panel").get_by_text("datalayer 2")).to_be_hidden()
    expect(peerB.locator(".panel").get_by_text("datalayer 2")).to_be_hidden()

    # Save delete to the server
    with peerA.expect_response(re.compile(".*/datalayer/delete/.*")):
        peerA.get_by_role("button", name="Save").click()
    expect(peerA.locator(".panel").get_by_text("datalayer 2")).to_be_hidden()
    expect(peerB.locator(".panel").get_by_text("datalayer 2")).to_be_hidden()


@pytest.mark.xdist_group(name="websockets")
def test_create_and_sync_map(new_page, asgi_live_server, tilelayer, login, user):
    # Create a syncable map with peerA
    peerA = login(user, prefix="Page A")
    peerA.goto(f"{asgi_live_server.url}/en/map/new/")
    peerA.get_by_role("button", name="Map advanced properties").click()
    expect(peerA.get_by_text("Real-time collaboration", exact=True)).to_be_hidden()
    with peerA.expect_response(re.compile("./map/create/.*")):
        peerA.get_by_role("button", name="Save Draft").click()
    peerA.get_by_role("button", name="Map advanced properties").click()
    expect(peerA.get_by_text("Real-time collaboration", exact=True)).to_be_visible()
    peerA.get_by_text("Real-time collaboration", exact=True).click()
    peerA.get_by_text("Enable real-time").click()
    peerA.get_by_role("button", name="Update permissions and editors").click()
    peerA.locator('select[name="share_status"]').select_option(str(Map.PUBLIC))
    with peerA.expect_response(re.compile("./update/settings/.*")):
        peerA.get_by_role("button", name="Save").click()
    expect(peerA.get_by_role("button", name="Cancel edits")).to_be_hidden()
    # Quit edit mode
    peerA.get_by_role("button", name="View").click()

    # Open map and go to edit mode with peer B
    peerB = new_page("Page B")
    peerB.goto(peerA.url)
    peerB.get_by_role("button", name="Edit").click()

    # Create a marker from peerA
    markersA = peerA.locator(".leaflet-marker-pane > div")
    markersB = peerB.locator(".leaflet-marker-pane > div")
    expect(markersA).to_have_count(0)
    expect(markersB).to_have_count(0)

    # Add a marker from peer A
    peerA.get_by_role("button", name="Edit").click()
    peerA.get_by_title("Draw a marker").click()
    peerA.locator("#map").click(position={"x": 220, "y": 220})
    expect(markersA).to_have_count(1)
    expect(markersB).to_have_count(1)

    # Make sure only one layer has been created on peer B
    peerB.get_by_role("button", name="Open browser").click()
    expect(peerB.locator("summary").get_by_text("Layer 1")).to_be_visible()
    peerB.get_by_role("button", name="Close").click()

    # Save and quit edit mode again
    with peerA.expect_response(re.compile("./datalayer/create/.*")):
        peerA.get_by_role("button", name="Save").click()
    peerA.get_by_role("button", name="View").click()
    expect(markersA).to_have_count(1)
    expect(markersB).to_have_count(1)
    peerA.wait_for_timeout(500)
    expect(markersA).to_have_count(1)
    expect(markersB).to_have_count(1)

    # Peer B should not be in state dirty
    expect(peerB.get_by_role("button", name="View")).to_be_visible()
    expect(peerB.get_by_role("button", name="Cancel edits")).to_be_hidden()

    # Add a marker from peer B
    peerB.get_by_title("Draw a marker").click()
    peerB.locator("#map").click(position={"x": 200, "y": 200})
    expect(markersB).to_have_count(2)
    expect(markersA).to_have_count(1)
    with peerB.expect_response(re.compile("./datalayer/update/.*")):
        peerB.get_by_role("button", name="Save").click()
    expect(markersB).to_have_count(2)
    expect(markersA).to_have_count(1)
    peerA.get_by_role("button", name="Edit").click()
    expect(markersA).to_have_count(2)
    expect(markersB).to_have_count(2)


@pytest.mark.xdist_group(name="websockets")
def test_saved_datalayer_are_not_duplicated(new_page, asgi_live_server, tilelayer):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()

    # Create one tab
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    # Create a new datalayer
    peerA.get_by_title("Manage layers").click()
    peerA.get_by_title("Add a layer").click()
    peerA.locator("#map").click(position={"x": 220, "y": 220})
    # Save layer to the server, so now the datalayer exist on the server AND
    # is still in the live operations of peer A
    with peerA.expect_response(re.compile(".*/datalayer/create/.*")):
        peerA.get_by_role("button", name="Save").click()

    # Now load the map from another tab
    peerB = new_page("Page B")
    peerB.goto(peerA.url)
    peerB.get_by_role("button", name="Open browser").click()
    expect(peerB.get_by_text("Layer 1")).to_be_visible()
    peerB.get_by_role("button", name="Edit").click()
    peerA.wait_for_timeout(300)  # Let the synchro roll on.
    expect(peerB.get_by_text("Layer 1")).to_be_visible()


@pytest.mark.xdist_group(name="websockets")
def test_should_sync_saved_status(new_page, asgi_live_server, tilelayer):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()

    # Create two tabs
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page("Page B")
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    # Create a new marker from peerA
    peerA.get_by_title("Draw a marker").click()
    peerA.locator("#map").click(position={"x": 220, "y": 220})

    # Peer A should be in dirty state
    expect(peerA.locator("body")).to_have_class(re.compile(".*umap-is-dirty.*"))

    # Peer B should not be in dirty state
    expect(peerB.locator("body")).not_to_have_class(re.compile(".*umap-is-dirty.*"))

    # Create a new marker from peerB
    peerB.get_by_title("Draw a marker").click()
    peerB.locator("#map").click(position={"x": 200, "y": 250})

    # Peer B should be in dirty state
    expect(peerB.locator("body")).to_have_class(re.compile(".*umap-is-dirty.*"))

    # Peer A should still be in dirty state
    expect(peerA.locator("body")).to_have_class(re.compile(".*umap-is-dirty.*"))

    # Save layer to the server from peerA
    with peerA.expect_response(re.compile(".*/datalayer/create/.*")):
        peerA.get_by_role("button", name="Save").click()

    # Peer B should not be in dirty state
    expect(peerB.locator("body")).not_to_have_class(re.compile(".*umap-is-dirty.*"))

    # Peer A should not be in dirty state
    expect(peerA.locator("body")).not_to_have_class(re.compile(".*umap-is-dirty.*"))


@pytest.mark.xdist_group(name="websockets")
def test_should_sync_line_on_escape(new_page, asgi_live_server, tilelayer):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()

    # Create two tabs
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page("Page B")
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    # Create a new marker from peerA
    peerA.get_by_title("Draw a polyline").click()
    peerA.locator("#map").click(position={"x": 220, "y": 220})
    peerA.locator("#map").click(position={"x": 200, "y": 200})
    peerA.locator("body").press("Escape")

    expect(peerA.locator("path")).to_have_count(1)
    expect(peerB.locator("path")).to_have_count(1)


@pytest.mark.xdist_group(name="websockets")
def test_should_sync_datalayer_clear(
    new_page, asgi_live_server, tilelayer, map, datalayer
):
    map.settings["properties"]["syncEnabled"] = True
    map.edit_status = Map.ANONYMOUS
    map.save()

    # Create two tabs
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page("Page B")
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    expect(peerA.locator(".leaflet-marker-icon")).to_have_count(1)
    expect(peerB.locator(".leaflet-marker-icon")).to_have_count(1)

    # Clear layer in peer A
    peerA.get_by_role("button", name="Manage layers").click()
    peerA.get_by_role("button", name="Edit", exact=True).click()
    peerA.locator("summary").filter(has_text="Advanced actions").click()
    peerA.get_by_role("button", name="Empty").click()
    expect(peerA.locator(".leaflet-marker-icon")).to_have_count(0)
    expect(peerB.locator(".leaflet-marker-icon")).to_have_count(0)

    # Undo in peer A
    peerA.locator(".edit-undo").click()
    expect(peerA.locator(".leaflet-marker-icon")).to_have_count(1)
    expect(peerB.locator(".leaflet-marker-icon")).to_have_count(1)


@pytest.mark.xdist_group(name="websockets")
def test_should_save_remote_dirty_datalayers(new_page, asgi_live_server, tilelayer):
    map = MapFactory(name="sync", edit_status=Map.ANONYMOUS)
    map.settings["properties"]["syncEnabled"] = True
    map.save()

    assert not DataLayer.objects.count()

    # Create two tabs
    peerA = new_page("Page A")
    peerA.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")
    peerB = new_page("Page B")
    peerB.goto(f"{asgi_live_server.url}{map.get_absolute_url()}?edit")

    # Create a new layer from peerA
    peerA.get_by_role("button", name="Manage layers").click()
    peerA.get_by_role("button", name="Add a layer").click()

    # Create a new layer from peerB
    peerB.get_by_role("button", name="Manage layers").click()
    peerB.get_by_role("button", name="Add a layer").click()

    # Save from peerA to the server
    counter = 0

    def on_response(response):
        nonlocal counter
        if "/datalayer/create/" in response.url:
            counter += 1
        # Wait for the two datalayer saves
        if counter == 2:
            return True
        return False

    with peerA.expect_response(on_response):
        peerA.get_by_role("button", name="Save").click()

    assert DataLayer.objects.count() == 2



# Generated by Django 4.2.2 on 2023-09-27 08:50

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0013_datalayer_edit_status"),
    ]

    operations = [
        migrations.AddField(
            model_name="map",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, default=django.utils.timezone.now
            ),
            preserve_default=False,
        ),
        # Did not find a way to provide default from a column in the field creation
        migrations.RunSQL("UPDATE umap_map SET created_at=modified_at;"),
    ]



# Generated by Django 5.1.6 on 2025-02-26 16:18

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0026_datalayer_modified_at_datalayer_share_status"),
    ]

    operations = [
        migrations.AddField(
            model_name="map",
            name="tags",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=200),
                blank=True,
                default=list,
                size=None,
            ),
        ),
    ]



# Generated by Django 4.2.8 on 2024-02-19 17:40

import uuid

from django.db import migrations, models

drop_index = """DO $$
BEGIN
    EXECUTE 'ALTER TABLE umap_datalayer DROP CONSTRAINT ' || (
        SELECT indexname
        FROM pg_indexes
        WHERE tablename = 'umap_datalayer' AND indexname LIKE '%pk%'
    );
END $$;
"""


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0017_migrate_to_openstreetmap_oauth2"),
    ]

    operations = [
        # Add the new uuid field
        migrations.AddField(
            model_name="datalayer",
            name="uuid",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, null=True, serialize=False
            ),
        ),
        # Generate UUIDs for existing records
        migrations.RunSQL(
            "UPDATE umap_datalayer SET uuid = gen_random_uuid()",
            reverse_sql=migrations.RunSQL.noop,
        ),
        # Remove the primary key constraint
        migrations.RunSQL(drop_index, reverse_sql=migrations.RunSQL.noop),
        # Drop the "id" primary key…
        migrations.AlterField(
            "datalayer", name="id", field=models.IntegerField(null=True, blank=True)
        ),
        # Rename "id" to "old id"
        migrations.RenameField(
            model_name="datalayer", old_name="id", new_name="old_id"
        ),
        # … to put it back on the "uuid"
        migrations.AlterField(
            model_name="datalayer",
            name="uuid",
            field=models.UUIDField(
                default=uuid.uuid4,
                editable=False,
                unique=True,
                primary_key=True,
                serialize=False,
            ),
        ),
        # When applying the migration backwards, we need to drop the pk index
        # Before addding a new one.
        migrations.RunSQL(migrations.RunSQL.noop, reverse_sql=drop_index),
    ]



# Generated by Django 4.2.2 on 2023-10-30 17:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0015_alter_pictogram_pictogram"),
    ]

    operations = [
        migrations.AddField(
            model_name="pictogram",
            name="category",
            field=models.CharField(blank=True, max_length=300, null=True),
        ),
    ]



# Generated by Django 4.2.2 on 2023-09-19 06:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("umap", "0012_datalayer_settings"),
    ]

    operations = [
        migrations.AddField(
            model_name="datalayer",
            name="edit_status",
            field=models.SmallIntegerField(
                choices=[
                    (0, "Inherit"),
                    (1, "Everyone"),
                    (2, "Editors only"),
                    (3, "Owner only"),
                ],
                default=0,
                verbose_name="edit status",
            ),
        ),
    ]



import pytest

from umap.utils import ConflictError, merge_features


def test_adding_one_element():
    assert merge_features(["A", "B"], ["A", "B", "C"], ["A", "B", "D"]) == [
        "A",
        "B",
        "C",
        "D",
    ]


def test_adding_elements():
    assert merge_features(["A", "B"], ["A", "B", "C", "D"], ["A", "B", "E", "F"]) == [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
    ]
    # Order does not count
    assert merge_features(["A", "B"], ["B", "C", "D", "A"], ["A", "B", "E", "F"]) == [
        "B",
        "C",
        "D",
        "A",
        "E",
        "F",
    ]


def test_adding_one_removing_one():
    assert merge_features(["A", "B"], ["A", "C"], ["A", "B", "D"]) == ["A", "C", "D"]


def test_removing_one():
    assert merge_features(["A", "B"], ["A", "B", "C"], ["A", "D"]) == ["A", "C", "D"]


def test_removing_same_element():
    # No added element (otherwise we cannot know if "new" elements are old modified
    # or old removed and new added).
    assert merge_features(["A", "B", "C"], ["A", "B"], ["A", "B"]) == [
        "A",
        "B",
    ]


def test_removing_changed_element():
    with pytest.raises(ConflictError):
        merge_features(["A", "B"], ["A", "C"], ["A"])


def test_changing_removed_element():
    with pytest.raises(ConflictError):
        merge_features(["A", "B"], ["A"], ["A", "C"])


def test_changing_same_element():
    with pytest.raises(ConflictError):
        merge_features(["A", "B"], ["A", "D"], ["A", "C"])
    # Order does not count
    with pytest.raises(ConflictError):
        merge_features(["A", "B", "C"], ["B", "D", "A"], ["A", "E", "B"])


def test_merge_with_ids_raises():
    # If reference doesn't have ids, but latest and incoming has
    # We need to raise as a conflict
    reference = [
        {"properties": {}, "geometry": "A"},
        {"properties": {}, "geometry": "B"},
    ]
    latest = [
        {"properties": {id: "100"}, "geometry": "A"},
        {"properties": {id: "101"}, "geometry": "B"},
    ]
    incoming = [
        {"properties": {id: "200"}, "geometry": "A"},
        {"properties": {id: "201"}, "geometry": "B"},
    ]

    with pytest.raises(ConflictError):
        merge_features(reference, latest, incoming)
